import {TreeList} from "./TreeList";
import Drawer from '@material-ui/core/Drawer';
import React from "react";
import {SplitPanel} from "./SplitPanel";
import clsx from "clsx";
import {makeStyles} from "@material-ui/core/styles";
import {useParams} from "react-router-dom";
import ChartRender from "./ChartRender";
import {APIContext} from "./callWrapper";
import MenuIcon from "@material-ui/icons/Menu";
import IconButton from "@material-ui/core/IconButton";

const drawerWidth = 240;

const useStyles = makeStyles((theme) => ({
  drawer: {
    width: drawerWidth,
    flexShrink: 0,
    whiteSpace: 'nowrap',
  },
  drawerOpen: {
    position: "relative",
    width: drawerWidth,
    transition: theme.transitions.create('width', {
      easing: theme.transitions.easing.sharp,
      duration: theme.transitions.duration.enteringScreen,
    }),
  },
  toolbar: {
    // display: 'flex',
    alignItems: 'center',
    justifyContent: 'flex-end',
    padding: theme.spacing(0, 1),
    display:"flex",
    "flex-direction":"column",
    "align-items": "flex-start",
    // necessary for content to be below app bar
    ...theme.mixins.toolbar,
  },
  drawerClose: {
    zIndex: 1200,
    position: "fixed",
    top: "340px",
    display: "block",
    transition: theme.transitions.create('width', {
      easing: theme.transitions.easing.sharp,
      duration: theme.transitions.duration.leavingScreen,
    }),
    overflowX: 'hidden',
    width: theme.spacing(7) + 1,
    [theme.breakpoints.up('sm')]: {
      width: theme.spacing(9) + 1,
    },
  },
  container: {
    position: "relative",
    right: "-32px",
    display: "flex",
    width: "100%",
    "flex-direction": "row",
  },
  dashboard: {
    "flex-grow": 1,
    "padding-right": "30px"
  },
  paper: {
    position: 'absolute',
    width: 400,
    backgroundColor: theme.palette.background.paper,
    border: '2px solid #000',
    boxShadow: theme.shadows[5],
    padding: theme.spacing(2, 4, 3),
  },
}));

export function DashboardRoute({dimensions}) {
    const params = useParams()

    const [dashboardID, setDashboardID] = React.useState(params["dashboardID"])

    React.useEffect( () => {
        console.log("Updating component")
        setDashboardID(params["dashboardID"])
    }, [params])


    return <DashboardEditor key={dashboardID} dashboardID={dashboardID} windowSize={dimensions} />
}

export const DashboardEditor = ({windowSize, dashboardID}) => {

  const classes = useStyles();

  const containingRef = React.createRef()

  const client = React.useContext(APIContext)

  const [divWidth, setDivWidth] = React.useState()
  const [open, setOpen] = React.useState(true)
  const [tree, setTree] = React.useState(undefined )

  const saveDashboard = () => {
    client.modifyDashboard(dashboardID, {
      config: tree,
      // visibility: visibility,
    }).then( (response) => {
      console.log(response)
      //TODO: Success, notify user
    }).catch( (error) => {
      //TODO: Error, notify user
    })
  }

  const getUpdatedDivWidth = () => {
    if(containingRef.current) {
      const width = containingRef.current.getBoundingClientRect()["width"]
      //TODO: Is equals okay? Could slight variations trigger rerender loop? Might be fixed by delta
      if(width !== divWidth) {
        console.log(`Old ${divWidth}, new ${width}`)
        setDivWidth(width)
      }
    }
  }

  const resizeHandler = (event) => {
    console.log(event)
    if(event[0]?.contentRect?.width) {
      setDivWidth(event[0].contentRect.width)
    }
  }

  React.useEffect( () => {
    const resizeObserver = new ResizeObserver(resizeHandler)
    resizeObserver.observe(containingRef.current)
    client.getDashboard(dashboardID).then( (retrievedDashboard) => {
      setTree(JSON.parse(retrievedDashboard.config))
    }).catch( (error) => {
      //TODO: Error, dashboard not retrieved.
    })
    return () => {
      if(resizeObserver) {
        resizeObserver.disconnect()
      }

    }
  }, [])

  return (
    <div className={classes.container}>
      <div class={classes.dashboard} ref={containingRef}>
        {tree ? <SplitPanel tree={tree} onTreeChange={
          () => {

          }
        } parentSize={divWidth}/> : undefined}
      </div>
      <Drawer
          anchor={"right"}
          variant="permanent"
          className={clsx(classes.drawer, {
            [classes.drawerOpen]: open,
            [classes.drawerClose]: !open,
          })}
          classes={{
            paper: clsx({
              [classes.drawerOpen]: open,
              [classes.drawerClose]: !open,
            }),
          }}
        >
        <div class={classes.toolbar}>
          <TreeList tree={tree} setTree={setTree} onSave={saveDashboard} />
          <br />
        </div>
      </Drawer>
    </div>

  )

}