import React, {createRef, useEffect, useState} from 'react';
import { makeStyles } from '@material-ui/core/styles';
import './MainContent.css';
import {Generator} from "./Generator";

const styleObject = { //Properties common to all orientations
  divider: {
    margin: "1rem",
    background: "#808080",
    border: "2px solid #808080",
  },
  splitView: {
    position:"relative",
    height: "100%",
    width: "100%",
    display: "flex",
    "align-items": "flex-start",
  },
  dividerHitbox: {
    position:"relative",
    "align-self": "stretch",
    display: "flex",
    "align-items": "center",
  },
  pane : {
    position:"relative",
    "overflow-x": "auto", // TODO: Not optimal, should not be required. Better solution?
    "overflow-y": "auto",
    "align-self": "stretch",
    "align-items": "center",
  },
};


// Properties dependent on orientation
styleObject["verticalDivider"] = Object.assign({ //Split along the horizontal axis
  width: "2px",
  height: "100%",
}, styleObject["divider"])

styleObject["horizontalDivider"] = Object.assign({ //Split along the vertical axis
    height: "2px",
    width: "100%",
}, styleObject["divider"])


styleObject["horizontalSplitView"] = Object.assign({ //Split along the horizontal axis
    position: "relative",
    "flex-direction": "row",
}, styleObject["splitView"])

styleObject["verticalSplitView"] = Object.assign({ //Split along the vertical axis
    position: "relative",
    "flex-direction": "column",
}, styleObject["splitView"])

styleObject["verticalDividerHitbox"] = Object.assign({ //Split along the horizontal axis
    cursor: "col-resize",
}, styleObject["dividerHitbox"])

styleObject["horizontalDividerHitbox"] = Object.assign({ //Split along the vertical axis
    cursor: "row-resize",
}, styleObject["dividerHitbox"])

styleObject["secondPane"] = Object.assign({
    flex: 1,
}, styleObject["pane"])

const useStyles = makeStyles(styleObject);

const LeftPanel = ({ children, leftWidth, setLeftWidth, setChildUpdate}) => {

  const leftRef = createRef();

  React.useEffect(() => {
    if (leftRef.current) {
      if (!leftWidth) {
        setLeftWidth(leftRef.current?.clientWidth);
        return;
      }

      leftRef.current.style.width = `${leftWidth}px`;
      setChildUpdate(leftWidth)
    }
  }, [leftRef, leftWidth]);

  return <div ref={leftRef}>{children}</div>;
};
const TopPanel = ({ children, height, setHeight, setChildUpdate }) => {

  const ref = createRef();

  React.useEffect(() => {
    if (ref.current) {
      if (!height) {
        setHeight(ref.current?.clientHeight);
        return;
      }
      ref.current.style.height = `${height}px`;
      setChildUpdate(height)
    }
  }, [ref, height]);

  return <div ref={ref}>{children}</div>;
};

export const SplitPanel = ({ tree, parentDragging, parentSize, onTreeChange}) => {

  const horizontal = tree.horizontal


  const MIN_DIMENSION = [100,100];
  const index = horizontal ? 0 : 1; //x-y-lookup index

  const classes = useStyles()
  const splitPaneRef = createRef();
  const firstPaneRef = createRef();
  const secondPaneRef = createRef();

  const [size, setSize] = useState(undefined);
  const [childUpdate, setChildUpdate] = useState(undefined);
  const [sizeBeforeDrag, setSizeBeforeDrag] = useState(undefined);
  const [separatorPosition, setSeparatorPosition] = useState(undefined);
  const [dragging, setDragging] = useState(false);

  const onMouseDown = (e) => {
    setSeparatorPosition([e.clientX,e.clientY])
    setSizeBeforeDrag(size);
    setDragging(true);
  };

  const onTouchStart = (e) => {
    setSeparatorPosition([e.touches[0].clientX,e.touches[0].clientY])
    setSizeBeforeDrag(size);
    setDragging(true);
  };

  const setTreeAspectRatio = (aspect) => {
    tree.aspect = aspect
    onTreeChange()
  }

  const resizePanel = (newSize) => {

    //Size bigger than minimal size for this dimension?
    if (newSize < MIN_DIMENSION[index]) {
        setSize(MIN_DIMENSION[index]);
        return;
      }

    //Residual size for second panel bigger than minimal size for this dimension?
    if (horizontal && splitPaneRef.current && secondPaneRef.current) {
      const deltaSize = size - newSize
      const secondSize = secondPaneRef.current.clientWidth //Limiting size will always width, height is unbounded
      if (secondSize + deltaSize < MIN_DIMENSION[index]) {
        setSize(secondSize + size - MIN_DIMENSION[index]);
        return;
      }
    }

    setSize(newSize)

  }

  const onMove = (clientPosition) => {
    if (dragging && size && separatorPosition) {

      const newSize = sizeBeforeDrag + clientPosition[index] - separatorPosition[index];


      resizePanel(newSize);
    }
  };

  const onMouseMove = (e) => {
    if (dragging) e.preventDefault();
    onMove([e.clientX,e.clientY]);
  };

  const onTouchMove = (e) => {
    onMove([e.touches[0].clientX,e.touches[0].clientY]);
  };

  const onMouseUp = () => {
    setSeparatorPosition(undefined);
    setDragging(false);
  };

  const rescale = () => {
    const [w, s] = tree.aspect
    const actualWidth = splitPaneRef.current?.clientWidth
    if(actualWidth) {
      console.log("rescaling")
      const scalingFactor = actualWidth / w
      const scaledSize = scalingFactor * s
      setTreeAspectRatio([w * scalingFactor, scaledSize])
      resizePanel(scaledSize)
    }
  };

  useEffect(() => {
    if(!dragging) {
      if(size) {
        setTreeAspectRatio([splitPaneRef.current?.clientWidth, size])
      }
    }
  }, [dragging]);

  useEffect(() => {
    if(parentSize) {
      rescale()
    }
  }, [parentSize]);


  useEffect(() => {
    document.addEventListener('mousemove', onMouseMove);
    document.addEventListener('touchmove', onTouchMove);
    document.addEventListener('mouseup', onMouseUp);
    document.addEventListener('touchend', onMouseUp);

    return () => {
      document.removeEventListener('mousemove', onMouseMove);
      document.removeEventListener('touchmove', onTouchMove);
      document.removeEventListener('mouseup', onMouseUp);
      document.removeEventListener('touchend', onMouseUp);
    };
  });

  useEffect(() =>  {
    rescale()
  }, [])

  const generateChild = (key, ref) => {
    return tree[key]
    ? ( tree[key].split
      ? <SplitPanel tree={tree[key].split} parentDragging={dragging} parentSize={childUpdate} onTreeChange={onTreeChange} />
      : (tree[key].generatorName
          ? (<Generator generatorName={tree[key].generatorName} generatorArguments={tree[key].args} parentSize={childUpdate} parentRef={ref}/> ?? <div /> )
          : <div />
        )
      )
    : <div />
  }
  const first = generateChild("c1", firstPaneRef)
  const second = generateChild("c2", secondPaneRef)

  const firstPanel = horizontal
    ? (<LeftPanel leftWidth={size} setLeftWidth={resizePanel} setChildUpdate={setChildUpdate}>
        {first}
      </LeftPanel>)
    : (<TopPanel height={size} setHeight={resizePanel} parentSize={parentSize} setChildUpdate={setChildUpdate}>
        {first}
      </TopPanel>)
  const enclosingDivClass = horizontal ? classes.horizontalSplitView : classes.verticalSplitView;
  const dividerHitboxClass = horizontal ? classes.verticalDividerHitbox : classes.horizontalDividerHitbox;
  const dividerClass = horizontal ? classes.verticalDivider : classes.horizontalDivider;

  return (
    <div ref={splitPaneRef} className={enclosingDivClass}>
      <div className={classes.pane} ref={firstPaneRef}>
        {firstPanel}
      </div>
      <div className={dividerHitboxClass}
           onMouseDown={onMouseDown}
           onTouchStart={onTouchStart}
           onTouchEnd={onMouseUp}>
        <div className={dividerClass}/>
      </div>
      <div ref={secondPaneRef} className={classes.secondPane}>
        {second}
          </div>
    </div>
  )
};