import axios from "axios";
import React from "react";

class Client {
  constructor(baseURL, name) {
    this.client = axios.create({
      baseURL: baseURL,
      timeout: 15000,
    });
    this.tokenEOL = 0;
    this.name = name;
    this.user = null;
    this.baseURL = baseURL;
    this.intervalID = null;
    this.refreshIntervalThreshold = 0.25
    this.loginStatusListener = {};

    this.cacheLifetimeDelta = 30;
    this.invalidateCache();

    if (this.baseURL.endsWith('/')) {
      // remove trailing slash
      this.baseURL = this.baseURL.substr(0, this.baseURL.length - 1);
    }
    //Check if a session exists by trying to refresh the current cookie
    this.refreshAccessToken().catch( (ex) => {})

  }

  login(username, password) {
    const param = {email: username, password: password};
    // send a token request to backend
    return this.client({method: "POST", url: `/api/token/`, data: param, headers: {"Content-Type": "application/json"}})
      .then((response) => {
        // If login was successful, invalidate any cache (there shouldnt be any, but doing it twice does no damage)
        this.invalidateCache();
        // Get the token lifetime, start autorefresh and get the user thats logged in
        return this.processChangedToken(response)
      }).then((user) => {
        // Save logged in user and inform UI-Elements of login status change
        this.user = user
        this.fireListenerEvent(true)
      });
  }

  clearAutoRefresh() {
    if (this.intervalID) {
      clearInterval(this.intervalID)
    }
  }

  setAutoRefresh() {
    /* Starts a daemon to keep the session token renewed while the session is running */

    // If a daemon is already running, end it
    this.clearAutoRefresh();
    // Get the time that the token is still valid
    const millisecondsLeft = this.tokenEOL - Date.now()
    // Start an interval that sends refresh requests after a percentage of token life time as specified in refreshIntervalThreshold
    this.intervalID = setInterval(() => {
      this.refreshAccessToken();
    }, Math.ceil(millisecondsLeft * this.refreshIntervalThreshold));
  }

  processChangedToken(response) {
        //FIXME: token data should not be available, this requires changes to the backend
        //TODO: Get EOL data from response to avoid concurrency bugs
        const payload = parseJwt(response.data['token']);
        this.tokenEOL = payload['exp'] * 1000
        this.setAutoRefresh()
        return this.getCurrentUser()
  }

  invalidateCache() {
    /* Remove cached objects and set cache lifetime into the past*/
    this.cacheLifetime = 0;
    this.cache = {
      charts:{},
      datasources: {},
      dashboards: {},
      groups: null,
    };
  }

  updateCache(cacheObjects) {
    /* Try to update cache elements and delete items that are too old */

    // If cache is over its lifetime, invalidate and reset lifetime
    if(this.cacheLifetime < Date.now()) {
      this.invalidateCache();
      this.cacheLifetime = Date.now() + this.cacheLifetimeDelta * 1000;
    }
    // If there are charts in cacheObjects set cache
    if(cacheObjects.charts) {
      Object.assign(this.cache['charts'], cacheObjects.charts);
    }
    // If there are datasources in cacheObjects set cache
    if(cacheObjects.datasources) {
      this.cache['datasources'] = cacheObjects.datasources;
    }
    // If there are groups in cacheObjects set cache
    if(cacheObjects.groups) {
      this.cache['groups'] = cacheObjects.groups;
    }

    if(cacheObjects.dashboards) {
      this.cache['dashboards'] = cacheObjects.dashboards;
    }
  }

  logout() {
    // Forget login state
    this.tokenEOL = 0;
    this.user = null;
    // Try to destroy cached elements
    this.invalidateCache()
    // Try to invalidate cookie
    this.client({method:'POST',
      url: `/api/token/blacklist/`,
    }).then( () => {
      // Deactivate token auto-renew
      this.clearAutoRefresh();
      // Inform subscribed UI-Elements that the client is logged out
      this.fireListenerEvent(false);
    }).catch( (error) => {});

  }

  addLoginStatusListener(listenerName, listener) {
    this.loginStatusListener[listenerName] = listener;
  }

  removeLoginStatusListener(listenerName) {
    this.loginStatusListener[listenerName].pop();
  }

  fireListenerEvent(event) {
    // Send login state to all loginStatusListener
    for (const [name,listener] of Object.entries(this.loginStatusListener)) {
      listener(event);
    }
  }

  async createDatasourceFromFile(file, name, visibility=0, onSuccess, onError) {
    await this.assertLogin()
    toBase64(file, (event) => {
      const header_length = event.target.result.search("base64") + "base64,".length
      const param = {
        datasource_name: name,
        data: event.target.result.slice(header_length),
        visibility: visibility,
      };
      this.client({
        method: "POST",
        url: `/api/datasources`,
        data: param,
      })
        .then(onSuccess).catch(onError);
    }, onError);
  }

  async createDatasourceFromURL(url, name, visibility=0) {
    await this.assertLogin()
    const param = {
      datasource_name: name,
      url: url,
      visibility: visibility,
    };
    const response = await this.client({
      method: "POST",
      url: `/api/datasources`,
      data: param,
    })
    return response.data;
  }

  async getDatasources(createdBefore, createdAfter, modifiedBefore, modifiedAfter, allowCache=false) {
    const key = `${createdBefore}${createdAfter}${modifiedBefore}${modifiedAfter}`
    if(allowCache && this.cache.datasources[key] != undefined && this.cacheLifetime > Date.now()) {
      return this.cache.datasources[key];
    } else {
      const config = {method: "GET", url: `/api/datasources`}
      const response = await this.client(config);

      const cacheObject = {datasources : {}};
      cacheObject.datasources[key] = response.data
      this.updateCache(cacheObject)

      return response.data;
    }
  }

  async getDatasource(datasourceID) {
    const config = {method: "GET", url: `/api/datasources/${datasourceID}`}
    const response = await this.client(config);
    return response.data;
  }

  async updateDatasource(datasourceID, name, visibility) {
    await this.assertLogin();
    const param = {};
    if (name) {
      param["datasource_name"] = name
    }
    if (visibility) {
      param["visibility"] = visibility
    }
    this.invalidateCache()
    const response = await this.client({
      method: "PATCH",
      url: `/api/datasources/${datasourceID}`,
      data: param,
    })
    return response.data;
  }

  async deleteDatasource(datasourceID) {
    await this.assertLogin();
    this.invalidateCache()
    const response = await this.client({
      method: "DELETE",
      url: `/api/datasources/${datasourceID}`,
    })
    return response.data;
  }

  async getChartTypes(datasourceID) {
    const config = {method: "GET", url: `/api/datasources/${datasourceID}/charttypes`}
    const response = await this.client(config);
    return response.data;
  }

  async getCharts(chartType, createdBefore, createdAfter, modifiedBefore, modifiedAfter, allowCache=false) {
    const key = `${chartType}${createdBefore}${createdAfter}${modifiedBefore}${modifiedAfter}`
    if(allowCache && this.cache.charts[key] != undefined && this.cacheLifetime > Date.now()) {
      return this.cache.charts[key];
    } else {
      const param = {}
      if (chartType) {
        param['chart_type'] = chartType;
      }
      if (createdBefore) {
        param['creation_time__lte'] = createdBefore;
      }
      if (createdAfter) {
        param['creation_time__gte'] = createdAfter;
      }
      if (modifiedBefore) {
        param['modification_time__lte'] = modifiedBefore;
      }
      if (modifiedAfter) {
        param['modification_time__gte'] = modifiedAfter;
      }
      const config = {method: "GET", url: `/api/charts`, params: param}
      const response = await this.client(config);

      const cacheObject = {charts : {}};
      cacheObject.charts[key] = response.data
      this.updateCache(cacheObject)

      return response.data;
    }
  }

  async createChart(datasourceID, chartType, chartConfig, name, downloadable = false, visibility = 0) {
    await this.assertLogin();
    this.invalidateCache()
    const data = {
      datasource: datasourceID,
      chart_type: chartType,
      config: chartConfig,
      chart_name: name,
      downloadable: downloadable,
      visibility: visibility
    }
    const config = {method: "POST", url: `/api/charts`, data: data, timeout: 0}
    const response = await this.client(config);
    return response.data;
  }

  async getChart(chartID) {
    const config = {method: "GET", url: `/api/charts/${chartID}`}
    const response = await this.client(config);
    return response.data;
  }

  async modifyChart(chartID, modifications = {}) {
    await this.assertLogin();
    this.invalidateCache()
    const config = {method: "PATCH", url: `/api/charts/${chartID}`, data: modifications}
    const response = await this.client(config);
    return response.data;
  }

  async deleteChart(chartID) {
    await this.assertLogin();
    this.invalidateCache()
    const config = {method: "DELETE", url: `/api/charts/${chartID}`}
    const response = await this.client(config);
    return response.data;
  }

  async getChartCodeURL(chartID) {
    const config = {method: "HEAD", url: `/api/charts/${chartID}/code`, maxRedirects: 0}
    const response = await this.client(config);
    return response.request.responseURL
  }

  async getChartCode(chartID) {
    const config = {method: "GET", url: `/api/charts/${chartID}/code`}
    const response = await this.client(config);
    return response.data;
  }

  getChartConfigURL(chartID) {
    return `${this.baseURL}/api/charts/${chartID}/config`
  }

  async getChartConfig(chartID) {
    const config = {method: "GET", url: `/api/charts/${chartID}/config`}
    const response = await this.client(config);
    return response.data;
  }

  async getChartFile(chartID, filename) {
    const config = {method: "GET", url: `/api/charts/${chartID}/files/${filename}`}
    const response = await this.client(config);
    return response.data;
  }


  //TODO: Add cache for dashboard objects
  async getDashboards(createdBefore, createdAfter, modifiedBefore, modifiedAfter, allowCache=false) {
    const key = `${createdBefore}${createdAfter}${modifiedBefore}${modifiedAfter}`
    if(allowCache && this.cache.dashboards[key] != undefined && this.cacheLifetime > Date.now()) {
      return this.cache.dashboards[key];
    } else {
      const param = {}
      if (createdBefore) {
        param['creation_time__lte'] = createdBefore;
      }
      if (createdAfter) {
        param['creation_time__gte'] = createdBefore;
      }
      if (modifiedBefore) {
        param['modification_time__lte'] = createdBefore;
      }
      if (modifiedAfter) {
        param['modification_time__gte'] = createdBefore;
      }
      const config = {method: "GET", url: `/api/dashboards`, params: param}
      const response = await this.client(config);

      const cacheObject = {dashboards: {}};
      cacheObject.dashboards[key] = response.data
      this.updateCache(cacheObject)

      return response.data;
    }
  }

  async createDashboard(name, dashboardConfig, visibility = 0) {
    await this.assertLogin();
    this.invalidateCache()
    if(typeof(dashboardConfig) === 'object') {
      dashboardConfig = JSON.stringify(dashboardConfig)
    }

    const data = {
      config: dashboardConfig,
      name: name,
      visibility: visibility
    }
    const config = {method: "POST", url: `/api/dashboards`, data: data}
    const response = await this.client(config);
    return response.data;
  }

  async getDashboard(dashboardID) {
    const config = {method: "GET", url: `/api/dashboards/${dashboardID}`}
    const response = await this.client(config);
    return response.data;
  }

  async modifyDashboard(dashboardID, modifications = {}) {
    await this.assertLogin();
    this.invalidateCache()
    if(typeof(modifications["config"]) === 'object') {
      modifications["config"] = JSON.stringify(modifications["config"])
    }
    const config = {method: "PATCH", url: `/api/dashboards/${dashboardID}`, data: modifications}
    const response = await this.client(config);
    return response.data;
  }

  async deleteDashboard(dashboardID) {
    await this.assertLogin();
    this.invalidateCache()
    const config = {method: "DELETE", url: `/api/dashboards/${dashboardID}`}
    const response = await this.client(config);
    return response.data;
  }
  async getShares(type, id) {
    const config = {method: "GET", url: `/api/${type}/${id}/shared`}
    const response = await this.client(config);
    return response.data;
  }

  getChartShares(chartID) {
    return this.getShares('charts', chartID);
  }

  getDatasourceShares(datasourceID) {
    return this.getShares('datasources', datasourceID);
  }

  getDashboardShares(dashboardID) {
    return this.getShares('dashboards', dashboardID);
  }

  async modifyShare(type, method, id, users, groups) {
    await this.assertLogin();
    this.invalidateCache()
    const data = {
      users: users,
      groups: groups,
    };
    const config = {method: method, url: `/api/${type}/${id}/shared`, data: data};
    const response = await this.client(config);
    return response.data;
  }

  removeChartShare(id, users, groups) {
    this.invalidateCache()
    return this.modifyShare('charts', 'DELETE', id, users, groups);
  }

  addChartShare(id, users, groups) {
    this.invalidateCache()
    return this.modifyShare('charts', 'PATCH', id, users, groups);
  }

  removeDashboardShare(id, users, groups) {
    this.invalidateCache()
    return this.modifyShare('dashboards', 'DELETE', id, users, groups);
  }

  addDashboardShare(id, users, groups) {
    this.invalidateCache()
    return this.modifyShare('dashboards', 'PATCH', id, users, groups);
  }

  removeDatasourceShare(id, users, groups) {
    this.invalidateCache()
    return this.modifyShare('datasources', 'DELETE', id, users, groups);
  }

  addDatasourceShare(id, users, groups) {
    this.invalidateCache()
    return this.modifyShare('datasources', 'PATCH', id, users, groups);
  }

  async getGroups(allowCache=false) {
    await this.assertLogin();
    if(allowCache && this.cache.groups != undefined && this.cacheLifetime > Date.now()) {
      return this.cache.groups;
    }
    const config = {method: "GET", url: `/api/groups`}
    const response = await this.client(config);
    this.cache.groups = response.data;
    return response.data;
  }

  async addGroup(name, isPublic = false, groupMembers= [], groupAdmins= []) {
    await this.assertLogin();
    this.invalidateCache()
    const data = {
      name: name,
      is_public: isPublic,
      group_members: groupMembers,
      group_admins: groupAdmins,
    };
    const config = {method: "POST", url: `/api/groups`, data: data}
    const response = await this.client(config);
    return response.data;
  }

  async getGroup(groupID) {
    await this.assertLogin();
    const config = {method: "GET", url: `/api/groups/${groupID}`}
    const response = await this.client(config);
    return response.data;
  }

  async getMultipleGroups(groupIDs) {
    const groups =  []
    for(let index in groupIDs) {
      groups[index] = this.getGroup(groupIDs[index])
    }
    return await Promise.all(groups)
  }

  async setGroupVisibility(groupID, is_public) {
    await this.assertLogin();
    const config = {method: "PATCH", url: `/api/groups/${groupID}`, data:{'is_public':is_public}}
    const response = await this.client(config);
    return response.data;
  }

  async deleteGroup(groupID) {
    await this.assertLogin();
    this.invalidateCache()
    const config = {method: "DELETE", url: `/api/groups/${groupID}`}
    const response = await this.client(config);
    return response.data;
  }

  async getGroupMembers(groupID) {
    await this.assertLogin();
    const config = {method: "GET", url: `/api/groups/${groupID}/properties`}
    const response = await this.client(config);
    return response.data;
  }

  async addGroupMembers(groupID, admins, members) {
    await this.assertLogin();
    this.invalidateCache()
    const data = {
      group_members: members,
      group_admins: admins,
    };
    const config = {method: "PATCH", url: `/api/groups/${groupID}/properties`, data:data}
    const response = await this.client(config);
    return response.data;
  }

  async removeGroupMembers(groupID, admins, members) {
    await this.assertLogin();
    this.invalidateCache()
    const data = {
      group_members: members,
      group_admins: admins,
    };
    const config = {method: "DELETE", url: `/api/groups/${groupID}/properties`, data:data}
    const response = await this.client(config);
    return response.data;
  }

  async addUser(email, password, username, firstname, lastname) {
    const data = {
      email: email,
      password: password,
      username: username,
      first_name: firstname,
      last_name: lastname,
    }
    const config = {method: 'POST', url: `/api/user/register/`, data: data};
    const response = await this.client(config);
    return response.data;
  }

  async getCurrentUser() {
    await this.assertLogin();
    const config = {method: 'GET', url: `/api/user/me/`};
    const response = await this.client(config);
    return response.data;
  }

  async updateCurrentUser(firstname= null, lastname = null, display_name = null, public_profile = null, real_name = null) {
    await this.assertLogin();
    const data = {}
    if(firstname != null) {
      data['first_name'] = firstname
    }
    if(lastname != null) {
      data['last_name'] = lastname
    }
    if(display_name != null) {
      data['username'] = display_name
    }
    if(public_profile != null) {
      data['public_profile'] = public_profile
    }
    if(real_name != null) {
      data['real_name'] = real_name
    }
    const config = {method: 'PATCH', url: `/api/user/me/`, data: data};
    const response = await this.client(config);
    return response.data;
  }

  async changePassword(oldPassword, newPassword) {
    await this.assertLogin();
    const data = {
      'oldPassword' : oldPassword,
      'newPassword' : newPassword,
    }
    const config = {method: 'POST', url: `/api/user/me/password/`, data: data};
    const response = await this.client(config);
    return response.data;
  }

  async changeEmail(password, newEmail) {
    await this.assertLogin();
    const data = {
      'password' : password,
      'newEmail' : newEmail,
    }
    const config = {method: 'POST', url: `/api/user/me/email/`, data: data};
    const response = await this.client(config);
    return response.data;
  }

  async getUser(id) {
    await this.assertLogin();
    const config = {method: 'GET', url: `/api/user/id/${id}`};
    const response = await this.client(config);
    return response.data;
  }

    async getUserByName(name) {
    await this.assertLogin();
    const config = {method: 'GET', url: `/api/user/search/`, params: {name: name}};
    const response = await this.client(config);
    return response.data;
  }

  async getUsers(ids) {
    await this.assertLogin();
    const config = {method: 'POST', url: `/api/user/`, data: {users: ids}, };
    const response = await this.client(config);
    return response.data;
  }

  refreshAccessToken() {
    return this.client({
      method: "POST",
      url: `/api/token/refresh/`,
      headers: {"Content-Type": "application/json"},
      data: {},
    })
      .then((response) => {
        return this.processChangedToken(response)
      }).then((user) => {
        this.user = user
        this.fireListenerEvent(true)
      }).catch( (ex) => {
        this.logout()
        throw ex;
      })
  }

  checkLoginSync() {
    return this.tokenEOL > Date.now()
  }

  assertLogin(force_check = false) {
    if(force_check || !this.checkLoginSync()) {
      return this.refreshAccessToken()
    }
  }

  async verifyToken(token) {
    try {
      let config = {
        method: "POST",
        url: `/api/token/verify/`,
        headers: {"Content-Type": "application/json"}
      };
      config['data'] = token ? {token: token} : {}
      await this.client(config)
      return true;
    } catch {
      return false;
    }
  }
}

export const APIContext = React.createContext()

function toBase64(file, onSuccess, onError) {
  const  reader = new FileReader();
  reader.onload = onSuccess;
  reader.onerror = onError;
  reader.readAsDataURL(file);
}

function parseJwt(token) {
  var base64Payload = token.split('.')[1];
  var payload = Buffer.from(base64Payload, 'base64');
  return JSON.parse(payload.toString());
}

export function createClient(baseURL, name = "default") {
  return new Client(baseURL, name);
}