import os
import sys
from flask import Flask, request, redirect, render_template, send_from_directory

import pive.environment as environment
import pive.inputmanager as inputmanager

from threading import Lock
from uuid import uuid4

from traceback import print_exc
import shutil

FILE_BASE_PATH = "/tmp/demonstrator"
STATIC_BASE_PATH = "/opt/demonstrator/static"
TEMPLATE_BASE_PATH = "/opt/demonstrator/pive_demo/templates"

app = Flask(__name__)

app_lock = Lock()

filemap = {}

@app.route('/')
def root():
    return render_template('site.html.jinja')

@app.route('/submit', methods = ['POST'])
def submit_file():
    file_object = request.files['file']
    try:
        #TODO: Check if file is actually processable before persisting
        handle = add_file(file_object.filename)
        folder_path = os.path.join(FILE_BASE_PATH,handle)
        os.makedirs(folder_path, exist_ok = True) 
        with open(os.path.join(folder_path,file_object.filename), "wb") as fp:
            fp.write(file_object.stream.read())
        return redirect(f"/file/{handle}", code=303)
    except:
        print_exc(file=sys.stderr)
        return ("Error: Unprocessable Entity", 422)

@app.route('/file/<handle>', methods = ['GET'])
def file_selection(handle):
    #TODO: Check for parameters and draw 
    try:
        outputpath = os.path.join(FILE_BASE_PATH,handle,"output")
        os.makedirs(outputpath, exist_ok = True) 
        manager = inputmanager.InputManager(mergedata=False)
        env = environment.Environment(inputmanager=manager, outputpath=outputpath)
        supported = env.load(os.path.join(FILE_BASE_PATH, handle, filemap[str(handle)]))
        if supported:
            return render_template('site.html.jinja', supported=supported)
        else:
            return render_template('site.html.jinja', error_message="No graphs supported for this data")
    #TODO: Better error handling, dont leak information
    except Exception as e:
        print_exc(file=sys.stderr)
        return render_template('site.html.jinja', error_message="An error occured: {}".format(e))

@app.route('/render/<handle>', methods = ['POST'])
def render_selection(handle):
    try:
        payload = request.get_json()

        outputpath = os.path.join(FILE_BASE_PATH,handle,"output")
        manager = inputmanager.InputManager(mergedata=False)
        env = environment.Environment(inputmanager=manager, outputpath=outputpath)
        supported = env.load(os.path.join(FILE_BASE_PATH, handle, filemap[str(handle)]))
        #FIXME: Unchecked user input, possible security risk
        chart_type = payload["chart_type"]
        if chart_type in supported:
            chart = env.choose(chart_type)
            chart.load_from_dict(payload)
            chart.set_div_hook("single_graph_container")
            chart.set_dataset_url(f"/data/{handle}/{chart_type}")
            chart.set_html_template(os.path.join(TEMPLATE_BASE_PATH, 'params.jinja'))
            env.render(chart)
            code = env.render_code(chart)[0] # Index 0 is js only
            return code
        else:
            #FIXME: Should only happen if someone is sending invalid chart types
            return ("Error: Someone is messing with the API endpoint", 400)
    #TODO: Better error handling, dont leak information
    except Exception as e:
        print_exc(file=sys.stderr)
        return ("An error occured: {}".format(e), 400)

@app.route('/params/<handle>', methods = ['POST'])
def get_parameter_list(handle):
    try:
        payload = request.get_json()

        outputpath = os.path.join(FILE_BASE_PATH,handle,"output")
        manager = inputmanager.InputManager(mergedata=False)
        env = environment.Environment(inputmanager=manager, outputpath=outputpath)
        supported = env.load(os.path.join(FILE_BASE_PATH, handle, filemap[str(handle)]))
        #FIXME: Unchecked user input, possible security risk
        chart_type = payload["chart_type"]
        if chart_type in supported:
            chart = env.choose(chart_type)
            chart.set_div_hook("single_graph_container")
            chart.set_dataset_url(f"/data/{handle}/{chart_type}")
            chart.set_html_template(os.path.join(TEMPLATE_BASE_PATH, 'params.jinja'))
            env.render(chart)
            return send_from_directory(directory=os.path.join(FILE_BASE_PATH,handle,"output"),filename=f"{chart_type}.html")
        else:
            #FIXME: Should only happen if someone is sending invalid chart types
            return ("Error: Someone is messing with the API endpoint", 400)
    #TODO: Better error handling, dont leak information
    except Exception as e:
        print_exc(file=sys.stderr)
        return ("An error occured: {}".format(e), 400)

@app.route('/data/<handle>/<chart_type>', methods = ['GET'])
def get_data(handle,chart_type):
    try:
        return send_from_directory(directory=os.path.join(FILE_BASE_PATH,handle,"output"),filename=f"{chart_type}.json")
    #TODO: Better error handling, dont leak information
    except Exception as e:
        print_exc(file=sys.stderr)
        return ("An error occured: {}".format(e), 400)


@app.route('/favicon.ico',  methods = ['GET'])
def get_favicon():
    return get_static("favicon.ico")

@app.route('/static/<filename>',  methods = ['GET'])
def get_static(filename):
    try:
        return send_from_directory(directory=STATIC_BASE_PATH, filename=filename)
    #TODO: Better error handling, dont leak information
    except Exception as e:
        print_exc(file=sys.stderr)
        return ("An error occured: {}".format(e), 404)


def add_file(filename):
    handle = str(uuid4())
    with app_lock:
        filemap[handle] = filename
    return handle

if __name__ == '__main__':
    shutil.rmtree(FILE_BASE_PATH, ignore_errors=True)
    app.run(host="0.0.0.0",port=8080)
